@@ -169,10 +169,74 @@ application_execute(const struct application *app, const struct prompt *prompt,
       memcpy(unescaped, launch_prefix, launch_len);
       unescaped[launch_len] = ' ';
       execute_dest = unescaped + launch_len + 1;
+
       if (id != NULL) {
+          setenv("DESKTOP_ENTRY_ID", id, 1);
+          /* Keep FUZZEL_DESKTOP_FILE_ID for backward compatibility */
           setenv("FUZZEL_DESKTOP_FILE_ID", id, 1);
       } else {
-          LOG_WARN("No Desktop File ID, not setting FUZZEL_DESKTOP_FILE_ID");
+          LOG_WARN("No Desktop File ID, not setting DESKTOP_ENTRY_ID");
+      }
+
+      if (app != NULL) {
+          if (app->desktop_file_path != NULL) {
+              setenv("DESKTOP_ENTRY_PATH", app->desktop_file_path, 1);
+          }
+
+          if (app->action_id != NULL) {
+              setenv("DESKTOP_ENTRY_ACTION", app->action_id, 1);
+          }
+
+          if (app->original_name != NULL) {
+              char *name_utf8 = ac32tombs(app->original_name);
+              setenv("DESKTOP_ENTRY_NAME", name_utf8, 1);
+              free(name_utf8);
+          }
+
+          if (app->localized_name != NULL) {
+              char *localized_name_utf8 = ac32tombs(app->localized_name);
+              setenv("DESKTOP_ENTRY_NAME_L", localized_name_utf8, 1);
+              free(localized_name_utf8);
+          }
+
+          if (app->comment != NULL) {
+              char *comment_utf8 = ac32tombs(app->comment);
+              setenv("DESKTOP_ENTRY_COMMENT", comment_utf8, 1);
+              setenv("DESKTOP_ENTRY_COMMENT_L", comment_utf8, 1); /* TODO: distinguish localized */
+              free(comment_utf8);
+          }
+
+          if (app->icon.name != NULL) {
+              setenv("DESKTOP_ENTRY_ICON", app->icon.name, 1);
+          }
+
+          if (app->original_generic_name != NULL) {
+              char *generic_name_utf8 = ac32tombs(app->original_generic_name);
+              setenv("DESKTOP_ENTRY_GENERICNAME", generic_name_utf8, 1);
+              free(generic_name_utf8);
+          }
+
+          if (app->localized_generic_name != NULL) {
+              char *localized_generic_name_utf8 = ac32tombs(app->localized_generic_name);
+              setenv("DESKTOP_ENTRY_GENERICNAME_L", localized_generic_name_utf8, 1);
+              free(localized_generic_name_utf8);
+          }
+
+          if (app->action_name != NULL) {
+              char *action_name_utf8 = ac32tombs(app->action_name);
+              setenv("DESKTOP_ENTRY_ACTION_NAME", action_name_utf8, 1);
+              free(action_name_utf8);
+          }
+
+          if (app->localized_action_name != NULL) {
+              char *localized_action_name_utf8 = ac32tombs(app->localized_action_name);
+              setenv("DESKTOP_ENTRY_ACTION_NAME_L", localized_action_name_utf8, 1);
+              free(localized_action_name_utf8);
+          }
+
+          if (app->action_id != NULL && app->icon.name != NULL) {
+              setenv("DESKTOP_ENTRY_ACTION_ICON", app->icon.name, 1);
+          }
       }
     } else {
       unescaped = xmalloc(execute_len + 1);
@@ -342,10 +406,21 @@ applications_destroy(struct application_list *apps)
         free(app->comment);
         free(app->icon.name);
 
+        /* Free additional metadata fields */
+        free(app->desktop_file_path);
+        free(app->action_id);
+        free(app->original_name);
+        free(app->localized_name);
+        free(app->action_name);
+        free(app->localized_action_name);
+        free(app->original_generic_name);
+        free(app->localized_generic_name);
+
         tll_free_and_free(app->keywords, free);
         tll_free_and_free(app->categories, free);
 
         free(app->dmenu_input);
+        free(app->dmenu_match_nth);
 
         switch (app->icon.type) {
         case ICON_NONE:
